﻿using System;
using System.Collections.Generic;
using System.DirectoryServices;
using EleWise.ELMA.ComponentModel;
using EleWise.ELMA.LdapSample.Settings;
using EleWise.ELMA.Logging;
using EleWise.ELMA.Security.Managers;
using EleWise.ELMA.Security.Models;
using EleWise.ELMA.Security.Services;
using IUser = EleWise.ELMA.Security.IUser;

namespace EleWise.ELMA.LdapSample.Services
{
    /// <summary>
    /// Сервис работы с LDAP
    /// </summary>
    [Service]
    public class LdapService : ILdapService
    {

        #region Private

        private static readonly ILogger IntegrationLdapLog = Logger.GetLogger("ElmaIntegrationLdap");

        /// <summary>
        /// Получение строки поиска пользователя по доменному имени
        /// </summary>
        private string GetSearchStringDN(LdapExternalMembershipModuleSettings settings, string searchString)
        {
            if (string.IsNullOrWhiteSpace(settings.LdapParamLogin))
                throw new ArgumentNullException("LdapLogin");
            if (string.IsNullOrWhiteSpace(searchString))
                throw new ArgumentNullException("Login");

            var s = string.Format("({0}={1})", settings.LdapParamLogin, searchString);
            if (!string.IsNullOrWhiteSpace(settings.LdapAuthFilter))
                s = string.Format("(&{0}{1})", settings.LdapAuthFilter, s);

            return s;
        }

        /// <summary>
        /// Получение произвольной строки поиска
        /// </summary>
        private string GetSearchString(LdapExternalMembershipModuleSettings settings, string searchString) //TODO в приват
        {
            var s = "";

            if (!string.IsNullOrWhiteSpace(searchString))
            {
                if (!string.IsNullOrWhiteSpace(settings.LdapParamLogin))
                    s = string.Format("({0}=*{1}*)", settings.LdapParamLogin, searchString);
                else
                    throw new ArgumentNullException("LdapLogin");

                if (!string.IsNullOrWhiteSpace(settings.LdapParamName))
                    s = string.Format("(|({0}=*{1}*){2})", settings.LdapParamName, searchString, s);

                if (!string.IsNullOrWhiteSpace(settings.LdapParamSecond))
                    s = string.Format("(|({0}=*{1}*){2})", settings.LdapParamSecond, searchString, s);

                if (!string.IsNullOrWhiteSpace(settings.LdapEMail))
                    s = string.Format("(|({0}=*{1}*){2})", settings.LdapEMail, searchString, s);

                if (!string.IsNullOrWhiteSpace(settings.LdapTelephoneNumber))
                    s = string.Format("(|({0}=*{1}*){2})", settings.LdapTelephoneNumber, searchString, s);

                if (!string.IsNullOrWhiteSpace(settings.LdapParamMiddle))
                    s = string.Format("(|({0}=*{1}*){2})", settings.LdapParamMiddle, searchString, s);

                if (!string.IsNullOrWhiteSpace(settings.LdapAuthFilter))
                    s = string.Format("(&{0}{1})", settings.LdapAuthFilter, s);
            }
            else
            {
                if (!string.IsNullOrWhiteSpace(settings.LdapAuthFilter))
                    s = string.Format("{0}", settings.LdapAuthFilter);
            }

            return s;
        }

        /// <summary>
        /// Преобразование шаблона авторизации
        /// </summary>
        private string SetLoginTemplate(SearchResult searchResult, LdapExternalMembershipModuleSettings settings)
        {
            return searchResult.Properties[settings.LdapParamLogin][0].ToString();
        }

        /// <summary>
        /// Считывание данных о пользователе из результата поиска
        /// </summary>
        private Security.Models.IUser ReadUserParam(SearchResult searchResult, LdapExternalMembershipModuleSettings settings)
        {
            var user = UserManager.Instance.Create();

            //Пытаемся прочитать ФИО и Адрес электронной почты из LDAP
            try
            {
                user.UserName = SetLoginTemplate(searchResult, settings);
            }
            catch (Exception exception)
            {
                var error = string.Format("Login read fail. Param: {0}", settings.LdapParamLogin);
                IntegrationLdapLog.Debug(error, exception);
            }

            ReadUserParam(searchResult, settings, user);
            return user;
        }

        /// <summary>
        /// Считывание данных о пользователе из результата поиска
        /// </summary>
        private bool ReadUserParam(SearchResult searchResult, LdapExternalMembershipModuleSettings settings, Security.Models.IUser user)
        {
            var hasChanges = false;

            try
            {
                var firstName = searchResult.Properties[settings.LdapParamName][0].ToString();
                if (user.FirstName != firstName)
                {
                    user.FirstName = firstName;
                    hasChanges = true;
                }
            }
            catch (Exception ex)
            {
                IntegrationLdapLog.Debug(ex);
            }

            try
            {
                var lastName = searchResult.Properties[settings.LdapParamSecond][0].ToString();
                if (user.LastName != lastName)
                {
                    user.LastName = lastName;
                    hasChanges = true;
                }
            }
            catch (Exception ex)
            {
                IntegrationLdapLog.Debug(ex);
            }

            try
            {
                var middleName = searchResult.Properties[settings.LdapParamMiddle][0].ToString();
                if (user.MiddleName != middleName)
                {
                    user.MiddleName = middleName;
                    hasChanges = true;
                }
            }
            catch (Exception ex)
            {
                IntegrationLdapLog.Debug(ex);
            }

            var fullName = user.GetFullName();
            if (user.FullName != fullName)
            {
                user.FullName = fullName;
                hasChanges = true;
            }

            

            try
            {
                var email = searchResult.Properties[settings.LdapEMail][0].ToString();
                if (user.EMail != email)
                {
                    user.EMail = email;
                    hasChanges = true;
                }
            }
            catch (Exception exception)
            {
                var error = string.Format("Email read fail. Param: {0}", settings.LdapEMail);
                IntegrationLdapLog.Debug(error, exception);
            }

            try
            {
                var telephoneNumber = searchResult.Properties[settings.LdapTelephoneNumber][0].ToString();
                if (user.WorkPhone != telephoneNumber)
                {
                    user.WorkPhone = telephoneNumber;
                    hasChanges = true;
                }
            }
            catch (Exception ex)
            {
                IntegrationLdapLog.Debug(ex);
            }

            return hasChanges;
        }

        /// <summary>
        /// Узел/объект LDAP
        /// </summary>
        private DirectoryEntry GetDirectoryEntry(LdapExternalMembershipModuleSettings settings)
        {
            if (settings == null)
                throw new Exception(SR.T("Отсутствуют настройки LDAP"));

            var directoryEntry = new DirectoryEntry
            {
                Path = "LDAP://" + settings.LdapUrl + "/" + settings.LdapPath,
                AuthenticationType = (AuthenticationTypes)settings.LdapAuthType,
                Username = settings.LdapLogin,
                Password = settings.LdapPassword
            };
            return directoryEntry;
        }

        /// <summary>
        /// Получение строки поиска для пользователя
        /// </summary>
        private string GetUserSearchString(LdapExternalMembershipModuleSettings settings, string searchString)
        {
            var s = "";

            if (!string.IsNullOrWhiteSpace(searchString))
            {
                if (!string.IsNullOrWhiteSpace(settings.LdapParamLogin))
                    s = string.Format("({0}={1})", settings.LdapParamLogin, searchString);
                else
                    throw new ArgumentNullException("LdapLogin");

                if (!string.IsNullOrWhiteSpace(settings.LdapAuthFilter))
                    s = string.Format("(&{0}{1})", settings.LdapAuthFilter, s);
            }

            return s;
        }

        #endregion

        #region Синхронизация пользователя

        ///<summary>
        /// Заполнение свойств пользователя данными из LDAP
        ///</summary>
        public void Sync(IUser user, LdapExternalMembershipModuleSettings settings)
        {
            if (user == null)
            {
                IntegrationLdapLog.Debug("Synchronization Error: user is null");
                return;
            }

            if (settings == null)
            {
                IntegrationLdapLog.Debug(SR.T("Отсутствуют настройки LDAP"));
                return;
            }

            if (string.IsNullOrWhiteSpace(settings.LdapUrl))
            {
                IntegrationLdapLog.Error(SR.T("Отсутствует адрес LDAP сервера"));
                return;
            }

            //Получение списка пользователей из LDAP по строке поиска
            using (var directoryEntry = GetDirectoryEntry(settings))
            {
                var searchString = GetUserSearchString(settings, user.UserName);
                if (string.IsNullOrWhiteSpace(searchString))
                {
                    IntegrationLdapLog.Debug(string.Format("Synchronization Error: can't read from LDAP {0}", user.UserName));
                    return;
                }

                using (var directorySearcher = new DirectorySearcher(directoryEntry, searchString))
                {
                    using (var searchResultCollection = directorySearcher.FindAll())
                    {
                        if (searchResultCollection != null && searchResultCollection.Count > 0)
                        {
                            ReadUserParam(searchResultCollection[0], settings, (Security.Models.IUser)user);
                        }
                        else
                            IntegrationLdapLog.Debug(SR.T("Не удалось получить дополнительную информацию о пользователе LDAP"));
                    }
                }
            }
        }

        #endregion

        #region Валидация

        /// <summary>
        /// Валидация по Distinguished Names
        /// </summary>
        public bool ValidatingDN(UserValidationContext context, LdapExternalMembershipModuleSettings settings)
        {
            return ValidatingDN(context.User.UserName, context.Password, settings);
        }

        /// <summary>
        /// Валидация по доменному имени
        /// </summary>
        public bool ValidatingTemplate(UserValidationContext context, LdapExternalMembershipModuleSettings settings)
        {
            return ValidatingTemplate(context.User.UserName, context.Password, settings);
        }

        /// <summary>
        /// Валидация по Distinguished Names
        /// </summary>
        public bool ValidatingDN(string userName, string password, LdapExternalMembershipModuleSettings settings)
        {
            if (settings == null || string.IsNullOrWhiteSpace(settings.LdapUrl))
            {
                IntegrationLdapLog.Debug(SR.T("Отсутствуют настройки LDAP"));
                return false;
            }

            if (settings == null)
                throw new Exception(SR.T("Отсутствуют настройки LDAP"));

            //ищем всех пользователей с указаным логином в OU и глубже взятом из "Путь к пользователям"
            using (var directoryEntry = new DirectoryEntry { Path = "LDAP://" + settings.LdapUrl + "/" + settings.LdapPath, AuthenticationType = (AuthenticationTypes)settings.LdapAuthType, Username = settings.LdapLogin, Password = settings.LdapPassword })
            {
                using (var directorySearcher = new DirectorySearcher(directoryEntry, GetSearchStringDN(settings, userName)))
                {
                    using (var searchResultCollection = directorySearcher.FindAll())
                    {
                        //такой пользователь должен быть только один
                        if (searchResultCollection == null)
                            throw new Exception("searchResultCollection == null");
                        if (searchResultCollection.Count == 0)
                            throw new Exception("User not found in LDAP");
                        if (searchResultCollection.Count > 1)
                            throw new Exception("It is more than one user in LDAP");

                        //получаем искомого пользователя
                        var searchResult = searchResultCollection[0];

                        if ((settings.LdapAuthType == (int)AuthenticationTypes.None || settings.LdapAuthType == (int)AuthenticationTypes.ReadonlyServer)
                            && string.IsNullOrEmpty(password))
                            return false;

                        //пытаемся получить объект, являющийся самим пользователем под логином-паролем пользователя
                        using (var lc = new DirectoryEntry { Path = searchResult.Path, AuthenticationType = (AuthenticationTypes)settings.LdapAuthType, Username = searchResult.Path.Remove(0, ((string)("LDAP://" + settings.LdapUrl + "/")).Length), Password = password })
                        {
                            var n = lc.NativeObject;
                        }
                        return true;
                    }
                }
            }
        }

        /// <summary>
        /// Валидация по доменному имени
        /// </summary>
        public bool ValidatingTemplate(string userName, string password, LdapExternalMembershipModuleSettings settings)
        {
            if (settings == null || string.IsNullOrWhiteSpace(settings.LdapUrl))
            {
                IntegrationLdapLog.Debug(SR.T("Отсутствуют настройки LDAP"));
                return false;
            }

            if ((settings.LdapAuthType == (int)AuthenticationTypes.None
                 || settings.LdapAuthType == (int)AuthenticationTypes.ReadonlyServer)
                && string.IsNullOrEmpty(password))
                return false;

            var ldapLogin = (settings.LdapAuthLogin == null)
                ? userName
                : settings.LdapAuthLogin.Replace("{$login}", userName);

            using (var lc = new DirectoryEntry { Path = "LDAP://" + settings.LdapUrl + "/" + settings.LdapPath, AuthenticationType = (AuthenticationTypes)settings.LdapAuthType, Username = ldapLogin, Password = password })
            {
                var n = lc.NativeObject;
            }

            return true;
        }

        #endregion

        #region Поиск

        /// <summary>
        /// Поиск пользователей в LDAP
        /// </summary>
        public IEnumerable<IUser> FindUsers(string searchString, LdapExternalMembershipModuleSettings settings, Guid serviceUid)
        {
            var result = new List<IUser>();

            try
            {
                if (searchString == null)
                    return result;

                if (settings == null)
                    throw new Exception(SR.T("Отсутствуют настройки LDAP"));
                if (string.IsNullOrWhiteSpace(settings.LdapUrl))
                    throw new Exception(SR.T("Отсутствует адрес LDAP сервера"));
                if (string.IsNullOrWhiteSpace(settings.LdapParamLogin))
                    throw new Exception(SR.T("Отсутствует настройка параметра \"Логин\""));


                //Получение списка пользователей из LDAP по строке поиска
                using (var directoryEntry = GetDirectoryEntry(settings))
                {
                    using (var directorySearcher = new DirectorySearcher(directoryEntry, GetSearchString(settings, searchString))
                    {
                        CacheResults = false
                    })
                    {
                        using (var searchResultCollection = directorySearcher.FindAll())
                        {
                            //Формирование результирующего списка IUser
                            foreach (SearchResult searchResult in searchResultCollection)
                            {
                                try
                                {
                                    var user = ReadUserParam(searchResult, settings);
                                    user.AuthProviderGuid = serviceUid;

                                    result.Add(user);
                                }
                                catch (Exception ex)
                                {
                                    var error = string.Format("{0}: {1}", "Import user from LDAP failed", ex.Message);
                                    IntegrationLdapLog.Error(error, ex);
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                var error = string.Format("{0}: {1}", SR.T("Не удалось получить список пользователей из LDAP"), ex.Message);
                IntegrationLdapLog.Debug(error, ex);
                throw new Exception(error);
            }

            return result;
        }

        #endregion

    }
}
